#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include "helper_cuda.h" // CUDA error checking

// integer ceiling division
#define DIV_CEIL(a,b) (((a)+(b)-1)/(b))
// maximum
#define MIN(a,b) (((a)<(b))?(a):(b))

#define BLOCK_SIZE 256
#define NUM_WARPS (BLOCK_SIZE/CUDA_WARP_SIZE)

__global__ void set(const uint N, int *values) {
	int gtid = blockIdx.x*blockDim.x + threadIdx.x;
	if (gtid<N)
		values[gtid] = (int) CUDA_WARP_SIZE*cosf((float)gtid);
}

__global__ void warp_scan1(const int *values, int *inclusive, int *exclusive) {
	__shared__ volatile int scan[NUM_WARPS*CUDA_WARP_SIZE];
	int tid = threadIdx.x;
	int lane = tid & CUDA_WARP_MASK;
	//int warp = tid / CUDA_WARP_SIZE;
	int gtid = blockIdx.x * blockDim.x + tid;

	// read from global
	int x = values[gtid];
	scan[tid] = x;

	int sum = x;
	for (int offset=1; offset<CUDA_WARP_SIZE; offset*=2) {
		if (lane>=offset) sum += scan[tid-offset];
		scan[tid] = sum;
	}
	inclusive[gtid] = sum;
	exclusive[gtid] = sum - x;
}

#define SCAN_STRIDE (CUDA_WARP_SIZE + CUDA_WARP_SIZE/2 + 1)

__global__ void warp_scan2(const int *values, int *inclusive, int *exclusive) {
	// Reserve a half warp of extra space.
	__shared__ volatile int scan[NUM_WARPS*SCAN_STRIDE];
	int tid = threadIdx.x;
	int lane = tid & CUDA_WARP_MASK;
	int warp = tid/CUDA_WARP_SIZE;
	int gtid = blockIdx.x*blockDim.x + tid;

	// Zero out the values at the start of the scan array and make a pointer
	// to scan + CUDA_WARP_SIZE / 2. Now s is addressable and zero for the first
	// sixteen values before the start of the array. This eliminates comparisons
	// and predicated execution.
	volatile int *s = scan + SCAN_STRIDE*warp + lane + CUDA_WARP_SIZE/2;
	s[-(CUDA_WARP_SIZE/2)] = 0;

	// Read from global memory.
	int x = values[gtid];
	s[0] = x;

	// Run each pass of the scan.
	int sum = x;
	for (int offset=1; offset<CUDA_WARP_SIZE; offset*=2) {
		// Since offset is at most 16, -offset will not read past the beginning
		// of the scan array.
		sum += s[-offset];
		s[0] = sum;
	}

	// Write sum to inclusive and sum - x (the original source element) to
	// exclusive.
	inclusive[gtid] = sum;
	exclusive[gtid] = sum - x;
}

__global__ void warp_scan3(const int *values, int *inclusive, int *exclusive) {
	int tid = threadIdx.x;
	int lane = tid & CUDA_WARP_MASK;
	//int warp = tid / CUDA_WARP_SIZE;
	int gtid = blockIdx.x * blockDim.x + tid;

	// read from global
	int x = values[gtid];

	int sum = x;
	#pragma unroll
	for (int i=1; i<CUDA_WARP_SIZE; i*=2) {
		int n = __shfl_up(sum, i, CUDA_WARP_SIZE);
		if (lane >= i) sum += n;
	}

	inclusive[gtid] = sum;
	exclusive[gtid] = sum - x;
}



int main(int argc, char **argv)
{
	int *d_values = NULL, *h_values = NULL; // pointers in Host and Device
	int *d_inclusive = NULL, *h_inclusive = NULL;
	int *d_exclusive = NULL, *h_exclusive = NULL;
	assert(2 == argc);
	const unsigned int N = 1<<atoi(argv[1]); // vector size
	assert(0 == N%CUDA_WARP_SIZE);

	checkCudaErrors(cudaMalloc(&d_values, N * sizeof(int)));
	checkCudaErrors(cudaMalloc(&d_inclusive, N * sizeof(int)));
	checkCudaErrors(cudaMalloc(&d_exclusive, N * sizeof(int)));
	checkCudaErrors(cudaMallocHost(&h_values, N * sizeof(int)));
	checkCudaErrors(cudaMallocHost(&h_inclusive, N * sizeof(int)));
	checkCudaErrors(cudaMallocHost(&h_exclusive, N * sizeof(int)));
	assert(d_values && h_values && d_inclusive && h_inclusive && d_exclusive && h_exclusive);

	dim3 block_size(BLOCK_SIZE);
	dim3 grid_size(DIV_CEIL(N,BLOCK_SIZE));
	set<<<grid_size, block_size>>>(N, d_values);
	getLastCudaError("set kernel failed");
	warp_scan3<<<grid_size, block_size>>>((const int*)d_values, d_inclusive, d_exclusive);
	getLastCudaError("warp_scan1 kernel failed");
	cudaDeviceSynchronize();

	checkCudaErrors(cudaMemcpy(h_values, d_values, N*sizeof(int), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_inclusive, d_inclusive, N*sizeof(int), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_exclusive, d_exclusive, N*sizeof(int), cudaMemcpyDefault));
	unsigned int i=0;
	unsigned int lane=0;
	for (i=0; i<N; i+=CUDA_WARP_SIZE) {
		int inclusive = 0, exclusive = 0;
		for (lane=0; lane<CUDA_WARP_SIZE; ++lane) {
			inclusive += h_values[i+lane];
			if (inclusive!=h_inclusive[i+lane] || exclusive!=h_exclusive[i+lane]) {
				goto EXIT; // "Go To Statement Considered Harmful"
			}
			exclusive += h_values[i+lane];
		}
	}
EXIT:
	if (N != i) { // if error, print complete lane.
		printf("Error in warp %d, lane %d\n", i/CUDA_WARP_SIZE, lane);
		for (unsigned int lane=0; lane<CUDA_WARP_SIZE; ++lane) {
			printf("%d ", h_values[i+lane]);
		} printf("\n");
		for (unsigned int lane=0; lane<CUDA_WARP_SIZE; ++lane) {
			printf("%d ", h_inclusive[i+lane]);
		} printf("\n");
		for (unsigned int lane=0; lane<CUDA_WARP_SIZE; ++lane) {
			printf("%d ", h_exclusive[i+lane]);
		} printf("\n");
	}

	checkCudaErrors(cudaFreeHost(h_values));
	checkCudaErrors(cudaFreeHost(h_inclusive));
	checkCudaErrors(cudaFreeHost(h_exclusive));
	checkCudaErrors(cudaFree(d_values));
	checkCudaErrors(cudaFree(d_inclusive));
	checkCudaErrors(cudaFree(d_exclusive));

	return 0;
}
